﻿/* WaterHeatPump.cs
 * 
 * Copyright (C) 2016 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

namespace Popolo.HVAC.HeatSource
{
  /// <summary>水熱源ヒートポンプ</summary>
  public class WaterHeatPump
  {

    #region 列挙型定義

    /// <summary>運転モード</summary>
    public enum OperatingMode
    {
      /// <summary>加熱運転</summary>
      Heating,
      /// <summary>冷却運転</summary>
      Cooling,
      /// <summary>停止</summary>
      ShutOff
    }

    #endregion

    #region 定数宣言

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>下限負荷率</summary>
    private const double MIN_PLOAD = 0.5;

    /// <summary>冷却運転の入力比特性係数</summary>
    private readonly double[] cAc = new double[5];
   
    /// <summary>加熱運転の入力比特性係数</summary>
    private readonly double[] cAh = new double[5];

    /// <summary>冷却運転の最大能力比特性係数</summary>
    private readonly double[] cBc = new double[] { 0.0313, -0.0100 };

    /// <summary>加熱運転の最大能力比特性係数</summary>
    private readonly double[] cBh = new double[] { 0.0247, -0.0023 };

    #endregion
    
    #region プロパティ
    
    /// <summary>運転モードを取得する</summary>
    public OperatingMode Mode { get; private set; }

    /// <summary>過負荷か否か</summary>
    public bool IsOverLoad { get; private set; }

    /// <summary>定格冷却能力[kW]を取得する</summary>
    public double NominalCoolingCapacity { get; private set; }

    /// <summary>冷水の定格質量流量[kg/s]を取得する</summary>
    public double NominalChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水の定格質量流量[kg/s]を取得する</summary>
    public double NominalCoolingWaterFlowRate { get; private set; }

    /// <summary>冷却運転時定格消費電力[kW]を取得する</summary>
    public double NominalCoolingEnergyConsumption { get; private set; }

    /// <summary>定格加熱能力[kW]を取得する</summary>
    public double NominalHeatingCapacity { get; private set; }

    /// <summary>温水の定格質量流量[kg/s]を取得する</summary>
    public double NominalHotWaterFlowRate { get; private set; }

    /// <summary>熱源水の定格質量流量[kg/s]を取得する</summary>
    public double NominalHeatSourceWaterFlowRate { get; private set; }

    /// <summary>加熱運転時定格消費電力[kW]を取得する</summary>
    public double NominalHeatingEnergyConsumption { get; private set; }

    /// <summary>最大冷却能力[kW]を取得する</summary>
    public double MaxCoolingCapacity { get; private set; }

    /// <summary>冷却負荷[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>冷水質量流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水質量流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>冷水出口温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterSetPoint { get; set; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    public double ChilledWaterInletTemperature { get; private set; }

    /// <summary>冷水出口温度[C]を取得する</summary>
    public double ChilledWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>最大加熱能力[kW]を取得する</summary>
    public double MaxHeatingCapacity { get; private set; }

    /// <summary>加熱負荷[kW]を取得する</summary>
    public double HeatingLoad { get; private set; }

    /// <summary>温水質量流量[kg/s]を取得する</summary>
    public double HotWaterFlowRate { get; private set; }

    /// <summary>熱源水質量流量[kg/s]を取得する</summary>
    public double HeatSourceWaterFlowRate { get; private set; }

    /// <summary>温水出口温度設定値[C]を設定・取得する</summary>
    public double HotWaterSetPoint { get; set; }

    /// <summary>温水入口温度[C]を取得する</summary>
    public double HotWaterInletTemperature { get; private set; }

    /// <summary>温水出口温度[C]を取得する</summary>
    public double HotWaterOutletTemperature { get; private set; }

    /// <summary>熱源水入口温度[C]を取得する</summary>
    public double HeatSourceWaterInletTemperature { get; private set; }

    /// <summary>熱源水出口温度[C]を取得する</summary>
    public double HeatSourceWaterOutletTemperature { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double EnergyConsumption { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP
    {
      get
      {
        if (Mode == OperatingMode.Cooling) return CoolingLoad / EnergyConsumption;
        else if (Mode == OperatingMode.Heating) return 0;
        else return 0;
      }
    }

    #endregion

    #region インスタンス変数

    /// <summary>定格条件における冷却運転時理論成績係数</summary>
    private readonly double copN_C;

    /// <summary>定格条件における加熱運転時理論成績係数</summary>
    private readonly double copN_H;

    /// <summary>定格条件における冷水出口温度[K]</summary>
    private readonly double tEvpoN_C;

    /// <summary>定格条件における冷却水出口温度[K]</summary>
    private readonly double tCndoN_C;

    /// <summary>定格条件における熱源水出口温度[K]</summary>
    private readonly double tEvpoN_H;

    /// <summary>定格条件における温水出口温度[K]</summary>
    private readonly double tCndoN_H;

    #endregion

    #region コンストラクタ

    /// <summary>インスタンスを初期化する</summary>
    /// <param name="coolingCapacity">冷却能力[kW]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="coolingEnergyConsumption">冷却運転時電力消費[kW]</param>
    /// <param name="heatingCapacity">加熱能力[kW]</param>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    /// <param name="heatSourceWaterFlowRate">熱源水質量流量[kg/s]</param>
    /// <param name="hotWaterOutletTemperature">温水出口温度[C]</param>
    /// <param name="heatSourceWaterInletTemperature">熱源水入口温度[C]</param>
    /// <param name="heatingEnergyConsumption">加熱運転時電力消費[kW]</param>
    public WaterHeatPump
      (double coolingCapacity, double chilledWaterFlowRate, double coolingWaterFlowRate,
      double chilledWaterOutletTemperature, double coolingWaterInletTemperature, double coolingEnergyConsumption,
      double heatingCapacity, double hotWaterFlowRate, double heatSourceWaterFlowRate,
      double hotWaterOutletTemperature, double heatSourceWaterInletTemperature, double heatingEnergyConsumption)
    {
      //冷却能力関連の情報初期化
      NominalCoolingCapacity = coolingCapacity;
      NominalCoolingEnergyConsumption = coolingEnergyConsumption;
      ChilledWaterSetPoint = chilledWaterOutletTemperature;
      tEvpoN_C = chilledWaterOutletTemperature + CONVERT_C_TO_K;
      tCndoN_C = (coolingCapacity + coolingEnergyConsumption) / (WATER_SPECIFIC_HEAT * coolingWaterFlowRate)
        + coolingWaterInletTemperature + CONVERT_C_TO_K;
      copN_C = tEvpoN_C / (tCndoN_C - tEvpoN_C);

      //加熱能力関連の情報初期化
      NominalHeatingCapacity = heatingCapacity;
      NominalHeatingEnergyConsumption = heatingEnergyConsumption;
      HotWaterSetPoint = hotWaterOutletTemperature;
      tEvpoN_H = - (heatingCapacity - heatingEnergyConsumption) / (WATER_SPECIFIC_HEAT * heatSourceWaterFlowRate)
        + heatSourceWaterInletTemperature + CONVERT_C_TO_K; 
      tCndoN_H = hotWaterOutletTemperature + CONVERT_C_TO_K;
      copN_H = tCndoN_H / (tCndoN_H - tEvpoN_H);

      //特性係数初期化
      double[] asc = new double[] { -0.7970, 1.0536, -0.2289, 1.2119, -0.2397 };
      double[] ash = new double[] { -0.0685, 0.1965, -0.0150, 1.0006, -0.1136 };
      double copC = NominalCoolingCapacity / NominalCoolingEnergyConsumption;
      double copH = NominalHeatingCapacity / NominalHeatingEnergyConsumption;
      for (int i = 0; i < asc.Length; i++)
      {
        cAc[i] = asc[i] / copC;
        cAh[i] = ash[i] / copH;
      }
    }

    #endregion

    #region publicメソッド

    /// <summary>冷却する</summary>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    public void CoolWater
      (double chilledWaterFlowRate, double coolingWaterFlowRate,
      double chilledWaterInletTemperature, double coolingWaterInletTemperature)
    {
      //負荷が無ければ停止
      double mcSpy = WATER_SPECIFIC_HEAT * chilledWaterFlowRate;
      CoolingLoad = mcSpy * (chilledWaterInletTemperature - ChilledWaterSetPoint);
      if (CoolingLoad <= 0)
      {
        ShutOff();
        return;
      }

      //冷却水出口温度を計算
      double mcSce = WATER_SPECIFIC_HEAT * coolingWaterFlowRate;
      double tScei = coolingWaterInletTemperature + CONVERT_C_TO_K;
      double tSpyi = chilledWaterInletTemperature + CONVERT_C_TO_K;
      double tSpyo = ChilledWaterSetPoint + CONVERT_C_TO_K;

      double tSceo, maxQSpy;
      bool isOverLoad;
      calcOutletTemperature
        (1, copN_C, tScei, tSpyi, ref tSpyo, tCndoN_C, tEvpoN_C, mcSce, mcSpy, CoolingLoad,
        NominalCoolingCapacity, cBc[1], cBc[0], cAc, out tSceo, out maxQSpy, out isOverLoad);
      if (isOverLoad) CoolingLoad = maxQSpy;

      //状態設定
      Mode = OperatingMode.Cooling;
      IsOverLoad = isOverLoad;
      ChilledWaterFlowRate = chilledWaterFlowRate;
      CoolingWaterFlowRate = coolingWaterFlowRate;
      ChilledWaterInletTemperature = chilledWaterInletTemperature;
      CoolingWaterInletTemperature = coolingWaterInletTemperature;
      MaxCoolingCapacity = maxQSpy;
      ChilledWaterOutletTemperature = tSpyo - CONVERT_C_TO_K;
      CoolingWaterOutletTemperature = tSceo - CONVERT_C_TO_K;
      EnergyConsumption = (mcSce * Math.Abs(tSceo - tScei)) - CoolingLoad;
      //加熱関連の情報初期化
      MaxHeatingCapacity = NominalHeatingCapacity;
      HeatingLoad = 0;
      HeatSourceWaterFlowRate = HotWaterFlowRate = 0.0;
      HeatSourceWaterOutletTemperature = HeatSourceWaterInletTemperature;
      HotWaterOutletTemperature = HotWaterInletTemperature;
    }

    /// <summary>加熱する</summary>
    /// <param name="hotWaterFlowRate">温水質量流量[kg/s]</param>
    /// <param name="heatSourceWaterFlowRate">熱源水質量流量[kg/s]</param>
    /// <param name="hotWaterInletTemperature">温水入口温度[C]</param>
    /// <param name="heatSourceWaterInletTemperature">熱源水入口温度[C]</param>
    public void HeatWater
      (double hotWaterFlowRate, double heatSourceWaterFlowRate,
      double hotWaterInletTemperature, double heatSourceWaterInletTemperature)
    {
      //負荷が無ければ停止
      double mcSpy = WATER_SPECIFIC_HEAT * hotWaterFlowRate;
      HeatingLoad = mcSpy * (HotWaterSetPoint - hotWaterInletTemperature);
      if (HeatingLoad <= 0)
      {
        ShutOff();
        return;
      }

      //冷却水出口温度を計算
      double mcSce = WATER_SPECIFIC_HEAT * heatSourceWaterFlowRate;
      double tScei = heatSourceWaterInletTemperature + CONVERT_C_TO_K;
      double tSpyi = hotWaterInletTemperature + CONVERT_C_TO_K;
      double tSpyo = HotWaterSetPoint + CONVERT_C_TO_K;

      double tSceo, maxQSpy;
      bool isOverLoad;
      calcOutletTemperature
        (-1, copN_H, tScei, tSpyi, ref tSpyo, tEvpoN_H, tCndoN_H, mcSce, mcSpy, HeatingLoad,
        NominalHeatingCapacity, cBh[0], cBh[1], cAh, out tSceo, out maxQSpy, out isOverLoad);
      if (isOverLoad) HeatingLoad = maxQSpy;

      //状態設定
      Mode = OperatingMode.Heating;
      IsOverLoad = isOverLoad;
      HotWaterFlowRate = hotWaterFlowRate;
      HeatSourceWaterFlowRate = heatSourceWaterFlowRate;
      HotWaterInletTemperature = hotWaterInletTemperature;
      HeatSourceWaterInletTemperature = heatSourceWaterInletTemperature;
      MaxHeatingCapacity = maxQSpy;
      HotWaterOutletTemperature = tSpyo - CONVERT_C_TO_K;
      HeatSourceWaterOutletTemperature = tSceo - CONVERT_C_TO_K;
      EnergyConsumption = HeatingLoad - (mcSce * Math.Abs(tSceo - tScei));
      //冷却関連の情報初期化
      MaxCoolingCapacity = NominalCoolingCapacity;
      CoolingLoad = 0;
      CoolingWaterFlowRate = ChilledWaterFlowRate = 0.0;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      ChilledWaterOutletTemperature = ChilledWaterInletTemperature;
    }
    
    /// <summary>熱源水側出口水温[K]を計算する</summary>
    /// <param name="cf">運転モード（冷却:+1,加熱:-1）</param>
    /// <param name="copTN">定格条件理論成績係数</param>
    /// <param name="tScei">熱源水側入口水温[K]</param>
    /// <param name="tSpyi">供給側入口水温[K]</param>
    /// <param name="tSpyo">供給側出口水温[K]</param>
    /// <param name="tSceoN">定格条件熱源水側出口水温[K]</param>
    /// <param name="tSpyoN">定格条件供給側出口水温[K]</param>
    /// <param name="mcSce">熱源水側熱容量流量[kW/K]</param>
    /// <param name="mcSpy">供給側熱容量流量[kW/K]</param>
    /// <param name="qSpy">供給熱量[kW]</param>
    /// <param name="qSpyN">定格条件供給熱量[kW]</param>
    /// <param name="bSce">最大能力補正係数1（熱源水側水温）</param>
    /// <param name="bSpy">最大能力補正係数2（供給側水温）</param>
    /// <param name="alpha">特性係数</param>
    /// <param name="tSceo">出力:熱源水側出口水温[K]</param>
    /// <param name="maxQSpy">出力:供給側最大能力[kW]</param>
    /// <param name="isOverLoad">出力:過負荷か否か</param>
    private static void calcOutletTemperature
      (int cf, double copTN, double tScei, double tSpyi, ref double tSpyo, double tSceoN, double tSpyoN,
      double mcSce, double mcSpy, double qSpy, double qSpyN, double bSce, double bSpy, double[] alpha,
      out double tSceo, out double maxQSpy, out bool isOverLoad)
    {
      //冷却水出口温度を計算
      double dsp = qSpyN / qSpy * (bSpy * (tSpyo - tSpyoN) - bSce * tSceoN + 1.0);
      double esp = qSpyN / qSpy * bSce;
      double fsp = cf * copTN / tSpyo;
      double asp = esp * (alpha[2] * esp + alpha[4] * fsp);
      double bsp = esp * (alpha[1] + 2 * alpha[2] * dsp) + alpha[3] * fsp
        + alpha[4] * (dsp * fsp - cf * copTN * esp) - mcSce / qSpy;
      double csp = alpha[0] + dsp * (alpha[1] + alpha[2] * dsp) 
        - cf * copTN * (alpha[3] + alpha[4] * dsp) + (cf * qSpy + mcSce * tScei) / qSpy;
      //解の存在確認
      bool isNanoLoad;
      double disc = bsp * bsp - 4 * asp * csp;
      if (0 < disc)
      {
        tSceo = 0.5 * (-bsp - Math.Sqrt(bsp * bsp - 4 * asp * csp)) / asp;
        maxQSpy = qSpyN * (1.0 + bSpy * (tSpyo - tSpyoN) + bSce * (tSceo - tSceoN));
        isOverLoad = maxQSpy < qSpy;  //過負荷判定
        isNanoLoad = (qSpy / maxQSpy) < MIN_PLOAD;
      }
      //極低負荷の場合
      else
      {
        tSceo = tScei;
        maxQSpy = qSpyN;
        isOverLoad = false;
        isNanoLoad = true;
      }

      //過負荷の場合
      if (isOverLoad)
      {
        double dff = -(cf * mcSpy + bSpy * qSpyN) / (bSce * qSpyN);
        double eff = 1d / bSce * (cf * mcSpy * tSpyi / qSpyN - 1.0 + bSpy * tSpyoN) + tSceoN;
        double fff = (alpha[3] + alpha[4]) * copTN;
        double gff = alpha[0] + alpha[1] + alpha[2] + cf;
        double bf = fff * (1 - dff);
        double aff = (bf - cf * gff) * mcSpy - dff * mcSce;
        double bff = (tSpyi * (-bf + cf * gff) - fff * eff) * mcSpy + (tScei - eff) * mcSce;
        double cff = fff * eff * mcSpy * tSpyi;
        tSpyo = 0.5 * (-bff - Math.Sqrt(bff * bff - 4 * aff * cff)) / aff;
        tSceo = dff * tSpyo + eff;
        maxQSpy = mcSpy * Math.Abs(tSpyi - tSpyo);
      }
      //極低負荷の場合
      else if (isNanoLoad)
      {
        double pl = 1d / MIN_PLOAD;
        double aplm = copTN * (alpha[3] + alpha[4] * pl);
        tSceo = (qSpy * (alpha[0] + (alpha[1] + alpha[2] * pl) * pl + cf * (1 - aplm)) + mcSce * tScei) / (mcSce - cf * qSpy * aplm / tSpyo);
        maxQSpy = qSpyN * (1.0 + bSpy * (tSpyo - tSpyoN) + bSce * (tSceo - tSceoN));
      }
    }

    /// <summary>機器を停止させる</summary>
    public void ShutOff()
    {
      Mode = OperatingMode.ShutOff;
      EnergyConsumption = 0.0;
      //冷却関連の情報初期化
      MaxCoolingCapacity = NominalCoolingCapacity;
      CoolingLoad = 0;
      CoolingWaterFlowRate = ChilledWaterFlowRate = 0.0;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature;
      ChilledWaterOutletTemperature = ChilledWaterInletTemperature;
      //加熱関連の情報初期化
      MaxHeatingCapacity = NominalHeatingCapacity;
      HeatingLoad = 0;
      HeatSourceWaterFlowRate = HotWaterFlowRate = 0.0;
      HeatSourceWaterOutletTemperature = HeatSourceWaterInletTemperature;
      HotWaterOutletTemperature = HotWaterInletTemperature;
    }

    #endregion

  }
}
